;This is a veeery basic DOM-Implementation
;But it should be enough for our purpose, this is not going to be a complete browser

Const DOM_TYPE_ELEMENT=1
Const DOM_TYPE_ATTR=2
Const DOM_TYPE_TEXT=3
Const DOM_TYPE_CDATA=4
Const DOM_TYPE_ENTITY=5
Const DOM_TYPE_ENTITYREF=6
Const DOM_TYPE_INSTRUCTION=7
Const DOM_TYPE_COMMENT=8
Const DOM_TYPE_DOCUMENT=9
;Further DOM-Node-Types not supported

Type DOM_Node
	Field name$
	Field value$ ;for attributes
	Field parent.DOM_Node
	Field typ
	
	Field attr.DOM_Node[64]
	Field attr_count
	
	Field childs.DOM_Node[128] ;not enough?
	Field childs_count
End Type


;Set of DOM-Functions on Elements, taken from: http://w3schools.com/dom/dom_element.asp
;For exact specification, see http://www.w3.org/2003/01/dom2-javadoc/org/w3c/dom/Node.html
;DEBUG - They are all NOT NULL-SAFE!

;DEBUG - not implemented yet:
;compareDocumentPosition()	Compares the document position of two nodes
;getAttributeNS()	Returns the value of an attribute (with a namespace)
;getAttributeNodeNS()	Returns an attribute node (with a namespace) as an Attribute Object
;getFeature(feature,version)	Returns a DOM Object which implements the specialized APIs of the specified feature And version
;getUserData(key)	Returns the Object associated To a key on a this node. The Object must First have been set To this node by calling setUserData with the same key
;hasAttributeNS()	Returns whether an element has any attributes matching a specified name And namespace
;isDefaultNamespace(URI)	Returns whether the specified namespaceURI is the Default
;isSupported(feature,version)	Returns whether a specified feature is supported on the element
;lookupNamespaceURI()	Returns the namespace URI matching a specified prefix
;lookupPrefix()	Returns the prefix matching a specified namespace URI
;normalize()	Puts all Text nodes underneath this element (including attributes) into a "normal" form where only structure (e.g., elements, comments, processing instructions, CDATA sections, And entity references) separates Text nodes, i.e., there are neither adjacent Text nodes nor empty Text nodes
;removeAttributeNS()	Removes a specified attribute (with a namespace)
;setUserData(key,Data,handler)	Associates an Object To a key on the element
;setAttributeNS()	Adds a New attribute (with a namespace)
;setAttributeNodeNS(attrnode)	Adds a New attribute node (with a namespace)
;setIdAttribute(name,isId)	If the isId property of the Attribute Object is True, this method declares the specified attribute To be a user-determined ID attribute
;setIdAttributeNS(uri,name,isId)	If the isId property of the Attribute Object is True, this method declares the specified attribute (with a namespace) To be a user-determined ID attribute
;setIdAttributeNode(idAttr,isId)	If the isId property of the Attribute Object is True, this method declares the specified attribute To be a user-determined ID attribute

;baseURI	Returns the absolute base URI of the element
;localName	Returns the Local part of the name of the element
;namespaceURI	Returns the namespace URI of the element
;prefix	Sets Or returns the namespace prefix of the element
;schemaTypeInfo	Returns the Type information associated with the element

Function DOM_appendChild(parent.DOM_Node, child.DOM_Node)	;Adds a New child node, To an element, as the Last child node
	If child\parent=parent Then
		;First remove the node, according to specification
		DOM_removeChild(parent, child)
	EndIf
	child\parent=parent
	parent\childs[parent\childs_count]=child
	parent\childs_count=parent\childs_count+1
End Function

;NodeList "Pfusch"
Function DOM_childNodes_item.DOM_Node(node.DOM_Node, index) ;Returns the node at the specified index in a NodeList
	Return node\childs[index]
End Function
Function DOM_childNodes_length(node.DOM_Node) ;Returns the number of nodes in a NodeList
	Return node\childs_count
End Function
Function DOM_attributes_item.DOM_Node(node.DOM_Node, index) ;Returns the node at the specified index in a NodeList
	Return Return node\attr[index]
End Function
Function DOM_attributes_length(node.DOM_Node) ;Returns the number of nodes in a NodeList
	Return node\attr_count
End Function

Function cloneNode.DOM_Node(node.DOM_Node, deep=True) ;Clones an element
	;Exact behavior following this specification:
	;http://www.w3.org/2003/01/dom2-javadoc/org/w3c/dom/Node.html#cloneNode_boolean_
	Local ret.DOM_Node=New DOM_Node
	
	ret\attr_count=node\attr_count
	For i=0 To ret\attr_count-1
		ret\attr[i]=cloneNode(node\attr[i], deep)
	Next
	
	If deep Then
		ret\childs_count=node\childs_count
		For i=0 To ret\childs_count-1
			ret\childs[i]=cloneNode(node\childs[i], deep)
		Next
	EndIf
	
	ret\name$=node\name$
	;ret\parent=node\parent
	ret\typ=node\typ
	ret\value$=node\value$
	Return ret
End Function

Function DOM_firstChild.DOM_Node(node.DOM_Node)	;Returns the First child of an element
	Return node\childs[0]
End Function

Function DOM_getAttribute$(node.DOM_Node, name$)	;Returns the specified attribute value of an element node
	name$=Lower(name$)
	For i=0 To node\attr_count-1
		If node\attr[i]\name$=name$ Then
			Return node\attr[i]\value$
		EndIf
	Next
	Return ""
End Function

Function DOM_getAttributeNode.DOM_Node(node.DOM_Node, name$)	;Returns the specified attribute node
	name$=Lower(name$)
	For i=0 To node\attr_count-1
		If node\attr[i]\name$=name$ Then
			Return node\attr[i]
		EndIf
	Next
End Function

Function DOM_hasAttribute(node.DOM_Node, name$) ;Returns True If an element has the specified attribute, otherwise False
	name$=Lower(name$)
	For i=0 To node\attr_count-1
		If node\attr[i]\name$=name$ Then
			Return True
		EndIf
	Next
	Return False
End Function

Function DOM_hasAttributes(node.DOM_Node) ;Returns True If an element has any attributes, otherwise False
	Return node\attr_count>0
End Function

Function DOM_hasChildNodes(node.DOM_Node) ;Returns True If an element has any child nodes, otherwise False
	Return node\childs_count>0
End Function

Function DOM_insertBefore(parent.DOM_Node, newNode.DOM_Node, refNode.DOM_Node) ;Inserts a New child node Before a specified, existing, child node
	Local pos=parent\childs_count
	;Search for refNode
	For i=0 To parent\childs_count-1
		If parent\childs[i]=refNode Then
			pos=i
			Exit
		EndIf
	Next
	If pos<>parent\childs_count Then
		newNode\parent=parent
		;Move all nodes from there
		For i=parent\childs_count-1 To pos Step -1
			parent\childs[i+1]=parent\childs[i]
		Next
		;Insert finally
		parent\childs[pos]=newNode
		parent\childs_count=parent\childs_count+1
	EndIf
End Function

Function DOM_isEqualNode(a.DOM_Node, b.DOM_Node) ;Checks If two elements are equal
	If a\typ<>b\typ Then Return False
	If a\parent<>b\parent Then Return False
	If a\name$<>b\name$ Then Return False
	If a\value$<>b\value$ Then Return False
	If a\attr_count<>b\attr_count Then Return False
	If a\childs_count<>b\childs_count Then Return False
	;DEBUG: Deep comparison of all childs and attrs???
	Return True
End Function

Function DOM_isSameNode(a.DOM_Node, b.DOM_Node) ;Checks If two elements are the same node
	Return a=b
End Function

Function DOM_lastChild.DOM_Node(node.DOM_Node) ;Returns the Last child of an element
	Return node\childs[node\childs_count-1]
End Function

Function DOM_nextSibling.DOM_Node(node.DOM_Node) ;Returns the Next node at the same node tree level
	Local n.DOM_Node=node\parent
	For i=0 To n\childs_count
		If n\childs[i]=node Then
			If i<n\childs_count-1 Then
				Return n\childs[i+1]
			Else
				Return Null
			EndIf
		EndIf
	Next
	Return Null
End Function

Function DOM_nodeName$(node.DOM_Node) ;Returns the name of an element
	Return node\name$
End Function

Function DOM_nodeType(node.DOM_Node) ;Returns the node Type of an element
	Return node\typ
End Function

Function DOM_nodeValue$(node.DOM_Node) ;Sets Or returns the value of an element
	Return DOM_get_nodeValue$(node)
End Function
Function DOM_set_nodeValue$(node.DOM_Node, value$)
	node\value$=value$
End Function
Function DOM_get_nodeValue$(node.DOM_Node)
	Return node\value$
End Function

Function DOM_ownerDocument.DOM_Node(node.DOM_Node) ;Returns the root element (document Object) For an element
	While node\parent<>Null
		node=node\parent
	Wend
	Return node
End Function

Function DOM_parentNode.DOM_Node(node.DOM_Node) ;Returns the parent node of an element
	Return node\parent
End Function

Function DOM_previousSibling.DOM_Node(node.DOM_Node) ;Returns the previous element at the same node tree level
	Local n.DOM_Node=node\parent
	For i=0 To n\childs_count
		If n\childs[i]=node Then
			If i>0 Then
				Return n\childs[i-1]
			Else
				Return Null
			EndIf
		EndIf
	Next
	Return Null
End Function

Function DOM_removeAttribute(node.DOM_Node, name$)	;Removes a specified attribute from an element
	name$=Lower(name$)
	Local pos=node\attr_count
	For i=0 To node\attr_count-1
		If node\attr[i]\name$=name$ Then
			pos=i
		EndIf
	Next
	If pos<>node\attr_count Then
		Delete node\attr[pos]
		For i=pos To node\attr_count-2
			node\attr[i]=node\attr[i+1]
		Next
		node\attr_count=node\attr_count-1
	EndIf
End Function

Function DOM_removeAttributeNode.DOM_Node(node.DOM_Node, name$)	;Removes a specified attribute node, And returns the removed node
	Local temp.DOM_Node=DOM_getAttributeNode(node.DOM_Node, name$)
	DOM_removeAttribute(node.DOM_Node, name$)
	Return temp
End Function

Function DOM_removeChild(node.DOM_Node, refNode.DOM_Node)	;Removes a child node from an element
	Local pos=node\childs_count
	;Search for refNode
	For i=0 To node\childs_count-1
		If node\childs[i]=refNode Then
			pos=i
			Exit
		EndIf
	Next
	If pos<>node\childs_count Then
		;Delete it
		Delete node\childs[pos]
		For i=pos To node\childs_count-2
			node\childs[i]=node\childs[i+1]
		Next
		node\childs_count=node\childs_count-1
	EndIf
End Function

Function DOM_replaceChild(parent.DOM_Node, newNode.DOM_Node, oldNode.DOM_Node)	;Replaces a child node in an element
	;Search for refNode
	For i=0 To parent\childs_count-1
		If parent\childs[i]=oldNode Then
			parent\childs[i]=newNode
			Exit
		EndIf
	Next
End Function

Function DOM_setAttribute(node.DOM_Node, name$, value$)	;Sets Or changes the specified attribute, To the specified value
	name$=Lower(name$)
	;Search for attr
	For i=0 To node\attr_count-1
		If node\attr[i]\name$=name$ Then
			node\attr[i]\value$=value$
		EndIf
	Next
	;Create new
	Local attr.DOM_Node=DOM_createAttribute(name$)
	attr\value$=value$
	DOM_setAttributeNode(node, attr)
End Function

Function DOM_setAttributeNode(node.DOM_Node, attr.DOM_Node)	;Sets Or changes the specified attribute node
	attr\parent=node
	node\attr[node\attr_count]=attr
	node\attr_count=node\attr_count+1
End Function

Function DOM_tagName(node.DOM_Node) ;Returns the tag name of an element
	Return node\name$
End Function

;element.textContent	Sets Or returns the textual content of a node And its descendants
Function DOM_textContent$(node.DOM_Node)
	Return DOM_get_textContent$(node)
End Function

Function DOM_get_textContent$(node.DOM_Node)
	Local ret$=""
	For i=0 To node\childs_count-1
		ret$=ret$+DOM_get_textContent$(node\childs[i])
		If node\childs[i]\typ=DOM_TYPE_TEXT Then
			ret$=ret$+node\childs[i]\value$
		EndIf
	Next
	Return ret$
End Function

Function DOM_set_textContent(node.DOM_Node, txt$)
	;Remove all Text-Nodes
	For i=0 To node\childs_count-1
		If node\childs[i]\typ=DOM_TYPE_TEXT Then
			DOM_removeChild(node,node\childs[i])
		EndIf
	Next
	;Set a new one
	Local n.DOM_Node=DOM_createTextNode(txt$)
	DOM_appendChild(node, n)
End Function

;Functions especially for document-Node
;See http://w3schools.com/dom/dom_document.asp

Function DOM_createAttribute.DOM_Node(name$)	;Creates an attribute node
	Local attr.DOM_Node=New DOM_Node
	attr\name$=name$
	attr\typ=DOM_TYPE_ATTR
	Return attr
End Function

Function DOM_createComment.DOM_Node(txt$)	;Creates a Comment node with the specified Text
	Local node.DOM_Node=New DOM_Node
	node\name$=txt$
	node\typ=DOM_TYPE_COMMENT
	Return node
End Function

Function DOM_createElement.DOM_Node(name$)	;Creates an Element node
	Local node.DOM_Node=New DOM_Node
	node\name$=name$
	node\typ=DOM_TYPE_ELEMENT
	Return node
End Function

Function DOM_createTextNode.DOM_Node(txt$)	;Creates a Text node
	Local node.DOM_Node=New DOM_Node
	node\name$="[TEXT]"
	node\value$=txt$
	node\typ=DOM_TYPE_TEXT
	Return node
End Function

Function DOM_getElementById.DOM_Node(node.DOM_Node, id$)	;Returns the element that has the ID attribute with the specified value
	Local temp$=DOM_getAttribute(node, "id")
	If temp$=id$ Then
		Return node
	EndIf
	For i=0 To node\childs_count-1
		n.DOM_Node=DOM_getElementById(node\childs[i], id$)
		If n<>Null Then
			Return n
		EndIf
	Next
	Return Null
End Function

Function DOM_getElementsByName(node.DOM_Node, name$)	;Accesses all elements with a specified name
	;In BB knnen wir keine Arrays zurck geben...
	;"Pfusch" mit XPath-Expression
	result=XPath_parseString("//"+name$, node)
	Return result
End Function

Function DOM_getElementsByTagName(node.DOM_Node, name$)	;Returns a NodeList containing all elements with the specified tagname
	Return DOM_getElementsByName(node.DOM_Node, name$)
End Function

Function DOM_renameNode(node.DOM_Node, name$)	;Renames the specified node
	node\name$=name$
End Function

;~IDEal Editor Parameters:
;~C#Blitz3D