
;Options
Global http_agent$="Mozilla/5.0 (Windows NT 6.1; rv:21.0) Gecko/20130401 Firefox/21.0"
Global http_saveFiles=False
Global http_saveCookies=False

;Internal variables
Global http_server$=""
Global http_stream=0
Global http_doc=0
Global http_writer=0

Type http_header
	Field name$
	Field value$
End Type

Type http_cookie
	Field name$
	Field value$
	Field server$
End Type

Function http_get(url$, filename$="")
	http_writer=0
	If http_doc<>0 Then FreeBank http_doc
	Delete Each http_header
	
	url$=http_prepareURL$(url$)
	
	DebugLog "GET ====================================="
	DebugLog "   Server: "+http_server$
	DebugLog "   File: "+url$
	
	http_loadCookies(http_server$)
	
	;Connection: Open new one or use keep-alive?
	If http_stream=0 Then
		http_stream=OpenTCPStream(http_server$,80)
	EndIf
	If http_stream=0 Then
		Return 0
	EndIf
	
	If filename$<>"" Then
		http_writer=WriteFile("download/"+filename$)
	Else
		If http_saveFiles Then
			http_writer=WriteFile("temp/"+Replace(Replace(url$,"/","_"),"?","_")+".txt")
		EndIf
	EndIf
	
	;Send Request
	WriteLine http_stream,"GET "+url$+" HTTP/1.1"
	WriteLine http_stream,"Host: "+http_server$
	WriteLine http_stream,"User-Agent: "+http_agent
	http_sendCookies()
	WriteLine http_stream,""
	
	Return http_readResponse()
End Function

Function http_post(url$, dat$)
	http_writer=0
	If http_doc<>0 Then FreeBank http_doc
	Delete Each http_header
	
	url$=http_prepareURL$(url$)
	
	DebugLog "POST ====================================="
	DebugLog "   Server: "+http_server$
	DebugLog "   File: "+url$
	DebugLog "   Data: "+dat$
	
	http_loadCookies(http_server$)
	
	;Connection: Open new one or use keep-alive?
	If http_stream=0 Then
		http_stream=OpenTCPStream(http_server$,80)
	EndIf
	If http_stream=0 Then
		Return 0
	EndIf
	
	If http_saveFiles Then http_writer=WriteFile("temp/"+Replace(Replace(url$,"/","_"),"?","_")+".txt")
	
	;Send Request
	WriteLine http_stream,"POST "+url$+" HTTP/1.1"
	WriteLine http_stream,"Host: "+http_server$
	WriteLine http_stream,"User-Agent: "+http_agent
	http_sendCookies()
	WriteLine http_stream,"Content-Length: "+Len(dat$)
	WriteLine http_stream,"Content-Type: application/x-www-form-urlencoded"
	WriteLine http_stream,""
	WriteLine http_stream,dat$
	WriteLine http_stream,""
	
	Return http_readResponse()
End Function

Function http_readResponse()
	;Read Response
	Local temp$, size
	Local txt$
	While txt$=""
		txt$=ReadLine(http_stream)
		If Eof(http_stream)<>0 Then
			DebugLog "Server closed connection without sending anything?"
			Return False
		EndIf
	Wend
	DebugLog "HTTP-Response "+txt$
	If Instr(txt$,"200")>0 Then
		;Read Headers
		Repeat
			txt$=ReadLine(http_stream)
			http_getHeader(txt$)
		Until txt$=""
		
		http_doc=CreateBank(0)
		usedHeader=False
		Repeat
			temp$=http_getHeaderbyName("Content-Length")
			If temp$="" Then
				temp$=ReadLine(http_stream)
				size=HexToInt(temp$)
			Else
				usedHeader=True
				size=temp$
			EndIf
			
			DebugLog_HTTP "Size: "+temp$+"="+size
			If temp$="0" Then Exit
			
			If size>0 Then
				bank=CreateBank(size)
				
				;Read Content
				For i=0 To size-1
					byte=ReadByte(http_stream)
					PokeByte(bank,i,byte)
					If http_writer<>0 Then WriteByte(http_writer,byte)
				Next
				
				http_doc=MergeBanks(http_doc,bank)
			EndIf
			If usedHeader Then Exit
		Forever
		
		;FlushStream(http_stream)
		
		;DEBUG - Connection-keep-alive does not work well...
		;connection$=http_getHeaderbyName("Connection")
		;If connection$="close" Then
			CloseTCPStream http_stream
			http_stream=0
		;EndIf
		
		If http_saveFiles Then CloseFile http_writer
		Return 1
	Else If Instr(txt$,"302")>0 Then
		;Redirect
		;Read Headers
		Repeat
			txt$=ReadLine(http_stream)
			http_getHeader(txt$)
		Until txt$=""
		
		http_flushStream()
		
		loc$=http_getHeaderbyName("Location")
		loc$=http_prepareURL$(loc$)
		DebugLog "Redirect to "+loc$
		
		;DEBUG - Connection-keep-alive does not work well...
		connection$=http_getHeaderbyName("Connection")
		If connection$="close" Then
			CloseTCPStream http_stream
			http_stream=0
		EndIf
		
		Return http_get(loc$)
	Else
		While Not Eof(http_stream)
			DebugLog_HTTP ReadLine(http_stream)
		Wend
		CloseTCPStream http_stream
		http_stream=0
		
		Return 0
	EndIf
End Function

;Helper-Functions

Function http_prepareURL$(url$)
	If Left(url$,1)="." Then
		;Relative URL, Server must be known
		If http_server$="" Then
			Return ""
		End If
		url$=Mid(url$,2)
	Else If Left(url$,1)="/" Then
		;Relative URL, Server must be known
		If http_server$="" Then
			Return ""
		End If
	Else
		;Split URL in Domain and File/Query
		Select Left(url$,7)
			Case "http://"
				url$=Mid(url$,8)
			Case "https:/"
				DebugLog "HTTP-Notice: HTTPS:// not supported!"
				url$=Mid(url$,9)
		End Select
		For i=1 To Len(url$)
			If Mid(url$,i,1)="/" Then
				http_server$=Left(url$,i-1)
				url$=Mid(url$,i)
				Exit
			EndIf
		Next
		If http_server$="" Then
			http_server$=url$
			url$="/"
		EndIf
	EndIf
	Return url$
End Function

Function http_getHeader(txt$)
	Local index=Instr(txt$,":")
	If index>0 Then
		Local header.http_header=New http_header
		header\name$=Lower(Left(txt$,index-1))
		header\value$=Trim(Mid(txt$,index+1))
		DebugLog_HTTP "Header "+header\name$+":"+header\value$
		If header\name$="set-cookie" Then
			http_getCookies(header\value$)
		EndIf
	EndIf
End Function

Function http_getHeaderbyName$(name$)
	name$=Lower(name$)
	Local header.http_header
	For header=Each http_header
		If header\name$=name$ Then
			Return header\value$
		EndIf
	Next
	Return ""
End Function

Function http_getCookies(txt$)
	Local cookie.http_cookie
	Local index=Instr(txt$,"=")
	If index>0 Then
		name$=Left(txt$,index-1)
		For cookie=Each http_cookie
			If cookie\name$=name$ Then
				Exit
			EndIf
		Next
		If cookie=Null Then
			cookie.http_cookie=New http_cookie
		EndIf
		cookie\name$=name$
		cookie\value$=Trim(Mid(txt$,index+1))
		index=Instr(cookie\value$,";")
		If index>0 Then
			cookie\value$=Left(cookie\value$,index-1)
		EndIf
		cookie\server$=http_server$
		DebugLog_HTTP "Cookie: "+cookie\name$+"="+cookie\value$
	EndIf
	If http_saveCookies Then
		Local writer=WriteFile("cookies/"+http_server$+".txt")
		For cookie=Each http_cookie
			If cookie\server$=http_server$ Then
				WriteLine writer, cookie\name$
				WriteLine writer, cookie\value$
			EndIf
		Next
		CloseFile writer
	EndIf
End Function

Function http_loadCookies(server$="")
	;Cookies saved on HD? Load just once!
	Local cookie.http_cookie
	Local ok=True
	For cookie=Each http_cookie
		If cookie\server$=http_server$ Then
			ok=False
			Exit
		EndIf
	Next
	If ok Then
		Local reader=ReadFile("cookies/"+server$+".txt")
		If reader>0 Then
			While Not Eof(reader)
				cookie=New http_cookie
				cookie\name$=ReadLine(reader)
				cookie\value$=ReadLine(reader)
				cookie\server$=http_server$
			Wend
			CloseFile reader
		EndIf
	EndIf
End Function

Function http_sendCookies()
	Local cookie.http_cookie
	Local txt$
	For cookie=Each http_cookie
		If cookie\server$=http_server$ Then
			If txt$="" Then
				txt$="Cookie: "
			Else
				txt$=txt$+"; "
			EndIf
			txt$=txt$+cookie\name$+"="+cookie\value$
		EndIf
	Next
	If txt$<>"" Then
		WriteLine http_stream,txt$
	EndIf
End Function

Function http_deleteCookies(server$="", deleteFromDisk=False)
	Local cookie.http_cookie
	For cookie=Each http_cookie
		If cookie\server$=server$ Or server$="" Then
			Delete cookie
		EndIf
	Next
	If deleteFromDisk Then
		If server$="" Then
			Local dir=ReadDir("cookies")
			Repeat
				temp$=NextFile(dir)
				If FileType("cookies/"+temp$)=1 Then
					DeleteFile "cookies/"+temp$
				EndIf
			Forever
			CloseDir dir
		Else
			DeleteFile "cookies/"+server$+".txt"
		EndIf
	EndIf
End Function

Function http_flushStream()
	DebugLog_HTTP "FlushStream"
	Local x=0
	Repeat
		x=Eof(http_stream)
		If x<>0 Then Exit
		ReadLine(http_stream)
	Forever
	If x=-1 Then
		;Error in Stream, Reset
		DebugLog_HTTP "Error, Reset Stream"
		CloseTCPStream http_stream
		http_stream=0
	EndIf
End Function
;~IDEal Editor Parameters:
;~C#Blitz3D