Const xmlParser_STATE_INIT=0
Const xmlParser_STATE_IN_TAG=1
Const xmlParser_STATE_CDATA=2
Const xmlParser_STATE_ATTR=3
Const xmlParser_STATE_IN_STRING=4
Const xmlParser_STATE_CLOSING=5
Const xmlParser_STATE_AUTOCLOSE=6
Const xmlParser_STATE_COMMENT=7
Const xmlParser_STATE_COMMENT_CLOSING=8

;How much should we try to "correct" invalid XML?
;xmlParser_correctLevel may be changed during execution!
Const xmlParser_CORRECT_NONE=0
Const xmlParser_CORRECT_SIMILAR_TAGS=%001 ;<a>...<a>..</a> -> Close first <a> when coming to second one
Const xmlParser_CORRECT_MISSING_TAG=%010 ;<div><span>Hallo</div> -> Close span automatically
Const xmlParser_CORRECT_FORMS=%100 ;<form><input...><input...><input...></form> -> Close ALL opened inputs
Const xmlParser_CORRECT_ALL=%111

Global xmlParser_correctLevel=xmlParser_CORRECT_ALL
;Global xmlParser_ignoreWhiteSpace=True

Global xmlParser_state=xmlParser_STATE_INIT
Global xmlParser_byte=0
Global xmlParser_temp$=""
Global xmlParser_currentNode.DOM_Node
Global xmlParser_currentQuote=0
Global xmlParser_commentState=0

Dim correctTag$(33)
.TagsToCorrect
Data "a","b","button","br","center","em","form","h1","h2","h3","h4","h5","h6"
Data "head","i","img","input","li","link","meta","p","script","small","strong","style"
Data "table","td","th","thead","title","tr","u","ul"
Restore TagsToCorrect
For i=0 To 32
	Read correctTag$(i)
Next

;Dim emptyTag$(12)
;.EmptyTags
;Data "area","base","basefont","br","col","frame","hr"
;Data "img","input","link","meta","param"
;Restore EmptyTags
;For i=0 To 11
;	Read emptyTag$(i)
;Next

Function xmlParser_loadFromBank.DOM_Node(bank, ignoreScript=False, saveTree$="")
	If bank=0 Then Return
	
	If ignoreScript Then
		bank=xmlParser_scriptPreprocessing(bank)
	EndIf
	
	xmlParser_state=xmlParser_STATE_INIT
	xmlParser_temp$=""
	
	DebugLog "XML-Parse [bank] ====================================="
	
	xmlParser_currentNode=New DOM_Node ;Root
	xmlParser_currentNode\typ=DOM_TYPE_DOCUMENT
	Local ret.DOM_Node=xmlParser_currentNode
	
	For i=0 To BankSize(bank)-1
		xmlParser_byte=PeekByte(bank,i)
		xmlParser_nextByte()
	Next
	
	If saveTree$<>"" Then
		xmlParser_saveTree(ret,saveTree$)
	EndIf
	
	Return ret
End Function

Function xmlParser_loadFromFile.DOM_Node(file$, ignoreScript=False, saveTree$="")
	Local reader=ReadFile(file$)
	If reader=0 Then Return
	
	DebugLog "XML-Parse "+file$+" ====================================="
	
	Local bank=CreateBank(FileSize(file$))
	i=0
	While Not Eof(reader)
		byte=ReadByte(reader)
		PokeByte(bank,i,byte)
		i=i+1
	Wend
	CloseFile reader
	
	If ignoreScript Then
		bank=xmlParser_scriptPreprocessing(bank)
	EndIf
	
	Local ret.DOM_Node=xmlParser_loadFromBank.DOM_Node(bank)
	
	If saveTree$<>"" Then
		xmlParser_saveTree(ret,saveTree$)
	EndIf
	
	FreeBank bank
	
	Return ret
End Function

;Compare to JS-method DOMParser.parseFromString()
Function xmlParser_loadFromString.DOM_Node(txt$, saveTree$="")
	Local bank=CreateBank(Len(txt$))
	For i=1 To Len(txt$)
		PokeByte(bank,i-1,Asc(Mid(txt$,i,1)))
	Next
	Local ret.DOM_Node=xmlParser_loadFromBank.DOM_Node(bank)
	
	If saveTree$<>"" Then
		xmlParser_saveTree(ret,saveTree$)
	EndIf
	Return ret
End Function

Function xmlParser_saveTree(node.DOM_Node, filename$)
	Local writer=WriteFile("DOM-Trees\"+filename$)
	If writer<>0 Then
		DebugNode(node,"",True,True,writer)
	EndIf
	CloseFile writer
End Function

;Little bit of overhead to remove script-tags before doing xml-parsing
;Because some sites do not use //<[CDATA[...
Function xmlParser_scriptPreprocessing(bank)
	Local pattern[7]
	pattern[0]=Asc("<")
	pattern[1]=Asc("s")
	pattern[2]=Asc("c")
	pattern[3]=Asc("r")
	pattern[4]=Asc("i")
	pattern[5]=Asc("p")
	pattern[6]=Asc("t")
	
	Local pattern2[9]
	pattern2[0]=Asc("<")
	pattern2[1]=Asc("/")
	pattern2[2]=Asc("s")
	pattern2[3]=Asc("c")
	pattern2[4]=Asc("r")
	pattern2[5]=Asc("i")
	pattern2[6]=Asc("p")
	pattern2[7]=Asc("t")
	pattern2[8]=Asc(">")
	
	Local mode=0
	;Local lines=0
	
	For i=0 To BankSize(bank)-1
		byte=PeekByte(bank,i)
		;If byte=10 Then lines=lines+1
		If mode=0 Then ;Search for "<script"
			For j=0 To 5
				If byte=pattern[j] Then
					If j=5 Then
						;Found <script
						mode=1
						closing=False
						i=i+6
						Exit
					EndIf
					byte=PeekByte(bank,i+j+1)
				Else
					Exit
				EndIf
			Next
		Else If mode=1 Then ;Search for > to get whole script-start-tag
			If byte=Asc("/") Then
				closing=i
			EndIf
			If byte=Asc(">") Then
				If closing=i-1 Then
					;Autoclose
					mode=0
				Else
					mode=2
				EndIf
			EndIf
		Else If mode=2 Then
			For j=0 To 8
				If byte=pattern2[j] Then
					If j=8 Then
						;Found </script>
						mode=0
						i=i+8
						Exit
					EndIf
					byte=PeekByte(bank,i+j+1)
				Else
					Exit
				EndIf
			Next
			If mode=2 Then
				PokeByte bank,i,Asc(" ") ;Blank out the whole script -.- 
			EndIf
		EndIf
	Next
	Return bank
End Function

Function xmlParser_nextByte()
	Select xmlParser_byte
		Case Asc("<")
			If xmlParser_state<>xmlParser_STATE_IN_STRING And xmlParser_state<>xmlParser_STATE_COMMENT_CLOSING Then
				xmlParser_gotText()
				xmlParser_state=xmlParser_STATE_IN_TAG
			Else
				xmlParser_anyChar()
			EndIf
		Case Asc(">")
			If xmlParser_state=xmlParser_STATE_IN_TAG Or xmlParser_state=xmlParser_STATE_CLOSING Then
				xmlParser_gotTagName()
			Else If xmlParser_state=xmlParser_STATE_AUTOCLOSE Then
					;DebugLog "Found Autoclose-Tag"
					;Workaround to use our "got*"-Function
				xmlParser_state=xmlParser_STATE_CLOSING
				xmlParser_temp$=xmlParser_currentNode\name$
				xmlParser_gotTagName()
			Else If xmlParser_state=xmlParser_STATE_COMMENT_CLOSING Then
				If xmlParser_commentState<=0 Then
					xmlParser_gotComment()
					xmlParser_state=xmlParser_STATE_CDATA
				EndIf
			EndIf
			If xmlParser_state<>xmlParser_STATE_IN_STRING And xmlParser_state<>xmlParser_STATE_COMMENT_CLOSING Then
				xmlParser_state=xmlParser_STATE_CDATA
			Else
				xmlParser_anyChar()
			EndIf
		Case Asc(" ")
			xmlParser_whitespace()
		Case 9
			xmlParser_whitespace()
		Case 0
			xmlParser_whitespace()
			
		Case Asc("/")
			If xmlParser_state=xmlParser_STATE_IN_TAG Then
				If xmlParser_temp$="" Then
					xmlParser_state=xmlParser_STATE_CLOSING
				Else
					xmlParser_gotTagName()
					xmlParser_state=xmlParser_STATE_AUTOCLOSE
				EndIf
			Else If xmlParser_state=xmlParser_STATE_ATTR Then
				xmlParser_state=xmlParser_STATE_AUTOCLOSE
			Else
				xmlParser_anyChar()
			EndIf
		Case Asc("!")
			If xmlParser_state=xmlParser_STATE_IN_TAG Then
				xmlParser_state=xmlParser_STATE_COMMENT
				xmlParser_commentState=0
			Else
				xmlParser_anyChar()
			EndIf
		Case Asc("-")
			If xmlParser_state=xmlParser_STATE_COMMENT Then
				If xmlParser_temp$="" Then
					xmlParser_commentState=xmlParser_commentState+1
					If xmlParser_commentState=2 Then
						xmlParser_state=xmlParser_STATE_COMMENT_CLOSING
					EndIf
				Else
					xmlParser_anyChar()
				EndIf
			Else If xmlParser_state=xmlParser_STATE_COMMENT_CLOSING Then
				xmlParser_commentState=xmlParser_commentState-1
				;DEBUG - Kommentar-Parsing noch nicht wasserdicht!!!
				;Gegen-Beispiel: <!-- ->->- -->
			Else
				xmlParser_anyChar()
			EndIf
		Case Asc("=")
			Select xmlParser_state
				Case xmlParser_STATE_ATTR
					xmlParser_gotAttrName()
				Default
					xmlParser_anyChar()
			End Select
			
		Case 34 ;"
			xmlParser_quote()
		Case Asc("'")
			xmlParser_quote()
			
		Case 10
				;linebreak
		Case 13
				;linebreak
		Default
			xmlParser_anyChar()
	End Select
End Function

Function xmlParser_anyChar()
	xmlParser_temp$=xmlParser_temp$+Chr(xmlParser_byte)
End Function

Function xmlParser_whitespace()
	Select xmlParser_state
		Case xmlParser_STATE_IN_TAG
			If xmlParser_gotTagName() Then
				xmlParser_state=xmlParser_STATE_ATTR
			EndIf
;		Case xmlParser_STATE_CDATA
;			If Not xmlParser_ignoreWhiteSpace Then
;				xmlParser_anyChar()
;			EndIf
		Default
			xmlParser_anyChar() ;DEBUG - wann kann man whitespaces ignorieren? nach <> und vor anderen zeichen?
	End Select
End Function

Function xmlParser_quote()
	Select xmlParser_state
		Case xmlParser_STATE_IN_STRING
			If xmlParser_currentQuote=xmlParser_byte Then
				;DebugLog "Found String: "+xmlParser_temp$
				xmlParser_gotAttrValue()
				xmlParser_state=xmlParser_STATE_ATTR
			Else
				xmlParser_anyChar()
			EndIf
		Case xmlParser_STATE_ATTR
			xmlParser_currentQuote=xmlParser_byte
			xmlParser_state=xmlParser_STATE_IN_STRING
		Default
			xmlParser_anyChar()
	End Select
End Function

;got*-Functions similar to a SAX-Style-Parser ;)

Function xmlParser_gotText()
	If CutSpaces$(xmlParser_temp$)<>"" Then
		Local n.DOM_Node=DOM_createTextNode(xmlParser_temp$)
		DOM_appendChild(xmlParser_currentNode, n)
		DebugLog_XML "Found Text:     "+xmlParser_temp$
	EndIf
	xmlParser_temp$=""
End Function

Function xmlParser_gotTagName()
	Local tagname$=Lower(xmlParser_temp$)
	If xmlParser_state=xmlParser_STATE_CLOSING Then
		attachToScript=False
		If xmlParser_currentNode\name$<>tagname$ Then
			If xmlParser_currentNode\parent=Null Then
				DebugLog "**** We have some troubles, stop parsing this!"
				Return False
			EndIf
			If xmlParser_currentNode\parent\name$=tagname$ Then
				;We assume that there is just 1 missing close-tag, correct it automatically
				If (xmlParser_correctLevel And xmlParser_CORRECT_MISSING_TAG) Then
					DebugLog_XMLError "Tag "+xmlParser_currentNode\name$+" closed (based on assumption that closing tag is missing)"
					xmlParser_currentNode=xmlParser_currentNode\parent
				EndIf
			Else
				;Some special cases which can occur if page is not valid XML
				If xmlParser_currentNode\name$="script" Then
					DebugLog_XMLError "**** Special Case Script-Child: ignored"
					attachToScript=True
				Else
					If tagname$="form" Then
						t$=xmlParser_currentNode\name$
						If t$="input" Or t$="label" Or t$="textarea" Or t$="button" Then
							If (xmlParser_correctLevel And xmlParser_CORRECT_FORMS) Then
								DebugLog_XMLError "No valid XHTML-Form - Closing all"
								;DEBUG - Der "missgebildete" DOM-Tree sollte noch entwirrt werden
								;Is das berhaupt mglich????
								While xmlParser_currentNode<>Null
									xmlParser_currentNode=xmlParser_currentNode\parent
									If xmlParser_currentNode\name$=tagname$ Then Exit
								Wend
							EndIf
						Else
							DebugLog_XMLError "**** Problem: "+xmlParser_currentNode\name$+" <> "+tagname$
						EndIf
					Else
						DebugLog_XMLError "**** Problem: "+xmlParser_currentNode\name$+" <> "+tagname$
					EndIf
				EndIf
			EndIf
		EndIf
		If Not attachToScript Then
			xmlParser_currentNode=xmlParser_currentNode\parent
			DebugLog_XML "Tag "+tagname$+" closed"
		EndIf
	Else
		If tagname$=xmlParser_currentNode\name$ Then
			If (xmlParser_correctLevel And xmlParser_CORRECT_SIMILAR_TAGS) Then
				;same Tags nested in each other?
				;Maybe this is a mistake (forgot to close last tag). Most likely in the following tags:
				For i=0 To 32
					If tagname$=correctTag$(i) Then
						xmlParser_currentNode=xmlParser_currentNode\parent
						Exit
					EndIf
				Next
			EndIf
		EndIf
		If xmlParser_currentNode\name$<>"script" Then
			DebugLog_XML "Tag "+tagname$+" / an Parent "+xmlParser_currentNode\name$
			;DebugNode(xmlParser_currentNode, "             ", False)
			Local n.DOM_Node=DOM_createElement(tagname$)
			DOM_appendChild(xmlParser_currentNode, n)
			xmlParser_currentNode=n
		Else
			Print "**** Problem: child in script-tag not allowed"
			xmlParser_temp$=""
			Return False
		EndIf
	EndIf
	xmlParser_temp$=""
	Return True
End Function

Function xmlParser_gotAttrName()
	Local n.DOM_Node=DOM_createAttribute(CutSpaces$(Lower(xmlParser_temp$)))
	DOM_setAttributeNode(xmlParser_currentNode, n)
	xmlParser_currentNode=n
	;DebugLog "Found Attr:      "+xmlParser_temp$
	xmlParser_temp$=""
End Function

Function xmlParser_gotAttrValue()
	xmlParser_currentNode\value$=xmlParser_temp$
	DebugLog_XML "   Attribute: "+xmlParser_currentNode\name$+"="+xmlParser_currentNode\value$
	xmlParser_currentNode=xmlParser_currentNode\parent
	xmlParser_state=xmlParser_STATE_ATTR
	xmlParser_temp$=""
End Function

Function xmlParser_gotComment()
	Local n.DOM_Node=DOM_createComment(xmlParser_temp$)
	DOM_appendChild(xmlParser_currentNode, n)
	DebugLog_XML "Found comment:   "+xmlParser_temp$
	xmlParser_temp$=""
End Function

;~IDEal Editor Parameters:
;~C#Blitz3D